# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import absolute_import

try:
    # for Python 3
    from collections.abc import Mapping
except ImportError:
    # for Python 2.7
    from collections import Mapping

import csv

from .. import data
from ..exceptions import MatchingIDNotFound


class BiMap(object):
    def __init__(self, class_A, class_B, converters=(int, int), filename=None):
        """
        Bi-bidirectional map class.

        Parameters
        ----------
        class_A, class_B: class types
            Input class types.
        converters: tuple, optional, default=(int,int)
            Converter functions applied on each entry (row) of the file
            providing the class_a-class_B matches.
            The order of the list elements must agree with that of the classes
            passed in in the constructor.
            By default, data on the file is assumed to be integers,
            which is the typical case for PDG IDs and alike.
        filename: string or file object, optional,
                  default='<class_B_name>_to_<class_A_name>.csv',
                  where the names are taken as lowercase.
            Specify a file from which to read all class_a-class_B matches.
            It does not matter whether the file contains the input as
            val_A,val_B or val_B,val_A.

        Examples
        --------
        >>> from particle import PDGID, PythiaID

        Basic/standard usage:
        >>> bimap = BiMap(PDGID, PythiaID)

        >>> bimap[PDGID(9010221)]
        <PythiaID: 10221>
        >>> bimap[PythiaID(10221)]
        <PDGID: 9010221>

        Advanced usage:
        >>> # Either pass a file name or a file object
        >>> from particle import data
        >>> filename = data.open_text(data, 'pdgid_to_pythiaid.csv')
        >>> bimap = BiMap(PDGID, PythiaID, filename=filename)
        """

        self.class_A = class_A
        self.class_B = class_B

        name_A = self.class_A.__name__.upper()
        name_B = self.class_B.__name__.upper()

        if filename is None:
            filename = "{a}_to_{b}.csv".format(a=name_A.lower(), b=name_B.lower())
            filename = data.open_text(data, filename)
        elif not hasattr(filename, "read"):
            # Conversion to handle pathlib on Python < 3.6:
            filename = str(filename)
            filename = open(filename)

        with filename as _f:
            self._to_map = {
                converters[1](v[name_B]): converters[0](v[name_A])
                for v in csv.DictReader(l for l in _f if not l.startswith("#"))
            }
            _f.seek(0)
            self._from_map = {
                converters[0](v[name_A]): converters[1](v[name_B])
                for v in csv.DictReader(l for l in _f if not l.startswith("#"))
            }

    def __getitem__(self, value):
        if isinstance(value, self.class_B):
            try:
                return self.class_A(self._to_map[value])
            except KeyError:
                pass
        elif isinstance(value, self.class_A):
            try:
                return self.class_B(self._from_map[value])
            except KeyError:
                pass
        msg = "Matching {a}-{b} for input {v} not found !".format(
            a=self.class_A.__name__, b=self.class_B.__name__, v=value
        )
        raise MatchingIDNotFound(msg)

    def __repr__(self):
        return "<{self.__class__.__name__}({a}-{b}): {n} matches>".format(
            self=self,
            a=self.class_A.__name__,
            b=self.class_B.__name__,
            n=self.__len__(),
        )

    def __str__(self):
        return repr(self)

    def __len__(self):
        """Returns the number of matches."""
        return len(self._to_map)


def DirectionalMaps(name_A, name_B, converters=(str, str), filename=None):
    """
        Directional map class providing a to and from mapping.

        Parameters
        ----------
        name_A, name_B: str
            Input names of information to be mapped.
        converters: tuple, optional, default=(str,str)
            Converter functions applied on each entry (row) of the file
            providing the name_a-name_B matches.
            The order of the list elements must agree with that of the
            object names passed in in the constructor.
            By default, data on the file is assumed to be strings,
            which is the typical case for particle names.
        filename: string or file object, optional, default='particle/data/conversions.csv'.
            Specify a file from which to read all name_a-name_B matches.
            It is assumed that the order of items in the file matches the order
            of arguments specified in the class constructor, hence val_A,val_B.

        Examples
        --------

        >>> from particle import data  # doctest: +SKIP
        >>> filename = data.open_text(data, 'a_to_b.csv')  # doctest: +SKIP
        >>> A2BMap, B2AMap = DirectionalMaps('A', 'B', filename=filename)  # doctest: +SKIP
        """

    name_A = name_A.upper()
    name_B = name_B.upper()

    fieldnames = None
    skipinitialspace = True

    if filename is None:
        filename = data.open_text(data, "conversions.csv")
    elif not hasattr(filename, "read"):
        # Conversion to handle pathlib on Python < 3.6:
        filename = str(filename)
        filename = open(filename)

    with filename as _f:
        to_map = {
            converters[1](v[name_B]): converters[0](v[name_A])
            for v in csv.DictReader(
                (l for l in _f if not l.startswith("#")),
                fieldnames=fieldnames,
                skipinitialspace=skipinitialspace,
            )
        }
        _f.seek(0)
        from_map = {
            converters[0](v[name_A]): converters[1](v[name_B])
            for v in csv.DictReader(
                (l for l in _f if not l.startswith("#")),
                fieldnames=fieldnames,
                skipinitialspace=skipinitialspace,
            )
        }

    return (
        DirectionalMap(name_A, name_B, from_map),
        DirectionalMap(name_B, name_A, to_map),
    )


class DirectionalMap(Mapping):
    def __init__(self, name_A, name_B, map):
        """
        Directional map class providing a A -> B mapping.

        Parameters
        ----------
        name_A, name_B: str
            Input names of information to be mapped.
        map: dict
            Input mapping as a dictionary.
        """

        self.name_A = name_A.upper()
        self.name_B = name_B.upper()

        self._map = map

    def __getitem__(self, value):
        try:
            return self._map[value]
        except KeyError:
            msg = "Matching {a}->{b} for input {v} not found !".format(
                a=self.name_A, b=self.name_B, v=value
            )
            raise MatchingIDNotFound(msg)

    def __iter__(self):
        return iter(self._map)

    def __repr__(self):
        return "<{self.__class__.__name__}({a}->{b}): {n} matches>".format(
            self=self, a=self.name_A, b=self.name_B, n=self.__len__()
        )

    def __str__(self):
        return repr(self)

    def __len__(self):
        """Returns the number of matches."""
        return len(self._map)



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

import sys

# Convenient access to the version number
from .version import version as __version__

# Direct access to PDGID
from .pdgid import PDGID

# Direct access to other ID classes
from .geant import Geant3ID
from .pythia import PythiaID

# Direct access to Particle (the CSV file is not read until a particle is accessed)
from .particle import Particle, SpinType, Parity, Charge, Inv, Status
from .particle import ParticleNotFound, InvalidParticle

# Direct access to Particle literals
from .particle import literals

sys.modules["particle.literals"] = literals

# Direct access to kinematics functions
from .particle import width_to_lifetime, lifetime_to_width

# Direct access to handy LaTeX to HTML particle name conversions
from .particle import latex_to_html_name

# Direct access to handy bi-directional maps
from .converters import Pythia2PDGIDBiMap



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Definitions of handy particle literals
======================================

The particle literals provide aliases for most common particles, with easily recognisable names.
The aliases are instances of the Particle class.

Typical use cases::

    >>> from particle import literals as lp
    >>> lp.pi_plus
    <Particle: name="pi+", pdgid=211, mass=139.57039 ± 0.00018 MeV>
    >>> lp.pi_plus.name
    'pi+'
    >>> from particle.literals import Lambda_b_0
    >>> Lambda_b_0
    <Particle: name="Lambda(b)0", pdgid=5122, mass=5619.60 ± 0.17 MeV>
    >>> Lambda_b_0.J
    0.5

List of available/defined particle literals:

{0}
"""

from ..shared_literals import common_particles
from .particle import Particle, ParticleNotFound

__doc = ""
for k, v in common_particles.items():
    try:
        locals()[k] = Particle.from_pdgid(v)
        __doc += "  {item!s} = Particle.from_pdgid({part})\n".format(item=k, part=v)
    except ParticleNotFound:
        pass

__doc__ = __doc__.format(__doc)


del Particle, ParticleNotFound, common_particles



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Helper (internal) module with common particle aliases.

See the particle.literals and the pdgid.literals submodules for the actually exposed aliases.
"""

from __future__ import absolute_import


# Names have to be valid identifiers.
# Values can be repeated.
common_particles = dict(
    # Gauge bosons
    photon=22,
    gamma=22,
    Z_0=23,
    W_plus=24,
    W_minus=-24,
    # Leptons
    e_minus=11,
    e_plus=-11,
    mu_minus=13,
    mu_plus=-13,
    tau_minus=15,
    tau_plus=-15,
    nu_e=12,
    nu_e_bar=-12,
    nu_mu=14,
    nu_mu_bar=-14,
    nu_tau=16,
    nu_tau_bar=-16,
    # Quarkonia
    Jpsi_1S=443,
    psi_2S=100443,
    eta_c_1S=441,
    eta_c_2S=100441,
    chi_c0_1P=10441,
    chi_c1_1P=20443,
    chi_c2_1P=445,
    Upsilon_1S=553,
    Upsilon_2S=100553,
    Upsilon_3S=200553,
    Upsilon_4S=300553,
    eta_b_1S=551,
    eta_b_2S=100551,
    chi_b0_1P=10551,
    chi_b1_1P=20553,
    chi_b2_1P=555,
    # Light mesons
    pi_0=111,
    pi_plus=211,
    pi_minus=-211,
    a_0_980_0=9000111,
    a_0_980_plus=9000211,
    rho_770_0=113,
    rho_770_plus=213,
    rho_770_minus=-213,
    a_1_1260_0=20113,
    a_1_1260_plus=20213,
    eta=221,
    etap_958=331,
    f_0_500=9000221,
    f_0_980=9010221,
    omega_782=223,
    phi_1020=333,
    K_L_0=130,
    K_S_0=310,
    K_plus=321,
    K_minus=-321,
    Kst_892_0=313,
    Kst_892_0_bar=-313,
    Kst_892_plus=323,
    Kst_892_minus=-323,
    K_0st_1430_0=10311,
    K_0st_1430_0_bar=-10311,
    K_0st_1430_plus=10321,
    K_0st_1430_minus=-10321,
    # Charm mesons
    D_0=421,
    D_0_bar=-421,
    D_plus=411,
    D_minus=-411,
    Dst_2007_0=423,
    Dst_2007_0_bar=-423,
    Dst_2010_plus=413,
    Dst_2010_minus=-413,
    D_s_plus=431,
    D_s_minus=-431,
    D_sst_plus=433,
    D_sst_minus=-433,
    D_s0st_2317_plus=10431,
    D_s0st_2317_minus=-10431,
    D_s1_2460_plus=20433,
    D_s1_2460_minus=-20433,
    D_s1_2536_plus=10433,
    D_s1_2536_minus=-10433,
    # Beauty mesons
    B_0=511,
    B_0_bar=-511,
    B_plus=521,
    B_minus=-521,
    B_s_0=531,
    B_s_0_bar=-531,
    # Beauty, charmed mesons
    B_c_plus=541,
    B_c_minus=-541,
    # Light baryons
    p=2212,
    proton=2212,
    p_bar=-2212,
    antiproton=-2212,
    n=2112,
    neutron=2112,
    n_bar=-2112,
    antineutron=-2112,
    Lambda=3122,
    Lambda_bar=-3122,
    Sigma_minus=3112,
    Sigma_plus_bar=-3112,
    Sigma_0=3212,
    Sigma_0_bar=-3212,
    Sigma_plus=3222,
    Sigma_minus_bar=-3222,
    Xi_minus=3312,
    Xi_plus_bar=-3312,
    Xi_0=3322,
    Xi_0_bar=-3322,
    Omega_minus=3334,
    Omega_plus_bar=-3334,
    Delta_1232_minus=1114,
    Delta_1232_plus_bar=-1114,
    Delta_1232_0=2114,
    Delta_1232_0_bar=-2114,
    Delta_1232_plus=2214,
    Delta_1232_minus_bar=-2214,
    Delta_1232_pp=2224,
    Delta_1232_mm_bar=-2224,
    Lambda_1520=3124,
    Lambda_1520_bar=-3124,
    # Charm baryons
    Lambda_c_plus=4122,
    Lambda_c_minus_bar=-4122,
    # Beauty baryons
    Lambda_b_0=5122,
    Lambda_b_0_bar=-5122,
    Xi_b_0=5232,
    Xi_b_0_bar=-5232,
    Xi_b_minus=5132,
    Xi_b_plus_bar=-5132,
    Omega_b_minus=5332,
    Omega_b_plus_bar=-5332,
)



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import absolute_import

from .pythia import Pythia2PDGIDBiMap
from .geant import Geant2PDGIDBiMap

from .evtgen import EvtGenName2PDGIDBiMap
from .evtgen import PDG2EvtGenNameMap, EvtGen2PDGNameMap



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
The Particle Data Group (PDG) defines the standard particle identification numbering scheme
in the form of a signed 7-digit number +/- N Nr Nl Nq1 Nq2 Nq3 Nj.

PDG IDs with more than 7 digits exist for non-standard particles such as Q-balls.
These follow outside the standard PDG numbering scheme.

This module provides the following:

- A pythonic version of the functions defined in HepPID and HepPDT,
  which work with PDG particle identification codes (PDG IDs).
- A few other functions extending the functionality of the HepXXX code.

References
----------

HepPDT and HepPID versions 3.04.01: http://lcgapp.cern.ch/project/simu/HepPDT/
"""

from __future__ import print_function, division, absolute_import

from enum import IntEnum
from typing import SupportsInt, Optional

PDGID_TYPE = SupportsInt


class Location(IntEnum):
    """
    Represents the location digit in the PDG numbering scheme, which provides a convenient index into the PID.
    """

    Nj = 1
    Nq3 = 2
    Nq2 = 3
    Nq1 = 4
    Nl = 5
    Nr = 6
    N = 7
    N8 = 8
    N9 = 9
    N10 = 10


def is_valid(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Is it a valid PDG ID?"""
    if _fundamental_id(pdgid) > 0:
        return True
    if is_meson(pdgid):
        return True
    if is_baryon(pdgid):
        return True
    if is_pentaquark(pdgid):
        return True
    if is_SUSY(pdgid):
        return True
    if is_Rhadron(pdgid):
        return True
    if is_dyon(pdgid):
        return True
    if is_diquark(pdgid):
        return True
    if is_pentaquark(pdgid):
        return True
    if _extra_bits(pdgid) > 0:
        if is_nucleus(pdgid):
            return True
        if is_Qball(pdgid):
            return True
        return False
    return False


def abspid(pdgid):
    # type: (PDGID_TYPE) -> int
    """Returns the absolute value of the PDG ID."""
    return abs(int(pdgid))


def is_lepton(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this PDG ID correspond to a lepton?"""
    if _extra_bits(pdgid) > 0:
        return False
    if _fundamental_id(pdgid) >= 11 and _fundamental_id(pdgid) <= 18:
        return True
    return False


def is_hadron(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this PDG ID correspond to a hadron?"""
    # Special case of proton and neutron:
    # needs to be checked first since _extra_bits(pdgid) > 0 for nuclei
    if abs(int(pdgid)) in (1000000010, 1000010010):
        return True
    if _extra_bits(pdgid) > 0:
        return False
    if is_meson(pdgid):
        return True
    if is_baryon(pdgid):
        return True
    if is_pentaquark(pdgid):
        return True
    return False


def is_meson(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this PDG ID correspond to a meson?"""
    if _extra_bits(pdgid) > 0:
        return False
    if abspid(pdgid) <= 100:
        return False
    if _fundamental_id(pdgid) <= 100 and _fundamental_id(pdgid) > 0:
        return False
    if abspid(pdgid) in (130, 210, 310):
        return True
    if abspid(pdgid) in (150, 350, 510, 530):
        return True
    if int(pdgid) in (110, 990, 9990):
        return True
    if (
        _digit(pdgid, Location.Nj) > 0
        and _digit(pdgid, Location.Nq3) > 0
        and _digit(pdgid, Location.Nq2) > 0
        and _digit(pdgid, Location.Nq1) == 0
    ):
        # check for illegal antiparticles
        if (
            _digit(pdgid, Location.Nq3) == _digit(pdgid, Location.Nq2)
            and int(pdgid) < 0
        ):
            return False
        else:
            return True
    return False


def is_baryon(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this PDG ID correspond to a baryon?"""
    if abspid(pdgid) <= 100:
        return False
    # Special case of proton and neutron:
    # needs to be checked first since _extra_bits(pdgid) > 0 for nuclei
    if abs(int(pdgid)) in (1000000010, 1000010010):
        return True
    if _extra_bits(pdgid) > 0:
        return False
    if _fundamental_id(pdgid) > 0 and _fundamental_id(pdgid) <= 100:
        return False
    if abspid(pdgid) == 2110 or abspid(pdgid) == 2210:
        return True
    if (
        _digit(pdgid, Location.Nj) > 0
        and _digit(pdgid, Location.Nq3) > 0
        and _digit(pdgid, Location.Nq2) > 0
        and _digit(pdgid, Location.Nq1) > 0
    ):
        return True
    if is_Rhadron(pdgid) or is_pentaquark(pdgid):
        return False
    return False


def is_diquark(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this PDG ID correspond to a diquark?"""
    if _extra_bits(pdgid) > 0:
        return False
    if abspid(pdgid) <= 100:
        return False
    if _fundamental_id(pdgid) <= 100 and _fundamental_id(pdgid) > 0:
        return False
    if (
        _digit(pdgid, Location.Nj) > 0
        and _digit(pdgid, Location.Nq3) == 0
        and _digit(pdgid, Location.Nq2) > 0
        and _digit(pdgid, Location.Nq1) > 0
    ):
        return True
    return False


def is_nucleus(pdgid):
    # type: (PDGID_TYPE) -> bool
    """
    Does this PDG ID correspond to a nucleus?

    Ion numbers are +/- 10LZZZAAAI.
    AAA is A - total baryon number
    ZZZ is Z - total charge
    L is the total number of strange quarks.
    I is the isomer number, with I=0 corresponding to the ground state.
    """
    # A proton can be a Hydrogen nucleus
    # A neutron can be considered as a nucleus when given the PDG ID 1000000010,
    # hence consistency demands that is_nucleus(neutron) is True
    if abspid(pdgid) in (2112, 2212):
        return True
    if _digit(pdgid, Location.N10) == 1 and _digit(pdgid, Location.N9) == 0:
        # Charge should always be less than or equal to the baryon number
        A_pdgid = A(pdgid)
        Z_pdgid = Z(pdgid)

        if A_pdgid is None or Z_pdgid is None:
            return False
        elif A_pdgid >= abs(Z_pdgid):
            return True
    return False


def is_pentaquark(pdgid):
    # type: (PDGID_TYPE) -> bool
    """
    Does the PDG ID correspond to a pentaquark?

    Pentaquark IDs are of the form +/- 9 Nr Nl Nq1 Nq2 Nq3 Nj, where Nj = 2J + 1 gives the spin
    and Nr Nl Nq1 Nq2 Nq3 denote the quark numbers in order Nr >= Nl >= Nq1 >= Nq2
    and Nq3 gives the antiquark number.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 9:
        return False
    if _digit(pdgid, Location.Nr) == 9 or _digit(pdgid, Location.Nr) == 0:
        return False
    if _digit(pdgid, Location.Nj) == 9 or _digit(pdgid, Location.Nl) == 0:
        return False
    if _digit(pdgid, Location.Nq1) == 0:
        return False
    if _digit(pdgid, Location.Nq2) == 0:
        return False
    if _digit(pdgid, Location.Nq3) == 0:
        return False
    if _digit(pdgid, Location.Nj) == 0:
        return False
    if _digit(pdgid, Location.Nq2) > _digit(pdgid, Location.Nq1):
        return False
    if _digit(pdgid, Location.Nq1) > _digit(pdgid, Location.Nl):
        return False
    if _digit(pdgid, Location.Nl) > _digit(pdgid, Location.Nr):
        return False
    return True


def is_Rhadron(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this PDG ID correspond to an R-hadron?

    An R-hadron is of the form 10abcdj, 100abcj, or 1000abj,
    where j = 2J + 1 gives the spin; b, c, and d are quarks or gluons;
    and a (the digit following the zero's) is a SUSY particle.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 1:
        return False
    if _digit(pdgid, Location.Nr) != 0:
        return False
    if is_SUSY(pdgid):
        return False
    # All R-hadrons have at least 3 core digits
    if (
        _digit(pdgid, Location.Nq2) == 0
        or _digit(pdgid, Location.Nq3) == 0
        or _digit(pdgid, Location.Nj) == 0
    ):
        return False
    return True


def is_Qball(pdgid):
    # type: (PDGID_TYPE) -> bool
    """
    Does this PDG ID correspond to a Q-ball or any exotic particle with electric charge beyond the qqq scheme?

    Ad-hoc numbering for such particles is +/- 100XXXY0, where XXX.Y is the charge.
    """
    if _extra_bits(pdgid) != 1:
        return False
    if _digit(pdgid, Location.N) != 0:
        return False
    if _digit(pdgid, Location.Nr) != 0:
        return False
    if (abspid(pdgid) // 10) % 10000 == 0:
        return False
    if _digit(pdgid, Location.Nj) != 0:
        return False
    return True


def is_dyon(pdgid):
    # type: (PDGID_TYPE) -> bool
    """
    Does this PDG ID correspond to a Dyon, a magnetic monopole?

    Magnetic monopoles and Dyons are assumed to have one unit of Dirac monopole charge
    and a variable integer number xyz units of electric charge,
    where xyz stands for Nq1 Nq2 Nq3.

    Codes 411xyz0 are used when the magnetic and electrical charge sign agree and 412xyz0 when they disagree,
    with the overall sign of the particle set by the magnetic charge.
    For now, no spin information is provided.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 4:
        return False
    if _digit(pdgid, Location.Nr) != 1:
        return False
    if _digit(pdgid, Location.Nl) != 1 and _digit(pdgid, Location.Nl) != 2:
        return False
    if _digit(pdgid, Location.Nq3) == 0:
        return False
    if _digit(pdgid, Location.Nj) != 0:
        return False
    return True


def is_SUSY(pdgid):
    # type: (PDGID_TYPE) -> bool
    """
    Does this PDG ID correspond to a SUSY particle?

    Fundamental SUSY particles have N = 1 or 2.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 1 and _digit(pdgid, Location.N) != 2:
        return False
    if _digit(pdgid, Location.Nr) != 0:
        return False
    if _fundamental_id(pdgid) == 0:
        return False
    return True


def has_down(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this particle contain a down quark?"""
    return _has_quark_q(pdgid, 1)


def has_up(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this particle contain an up quark?"""
    return _has_quark_q(pdgid, 2)


def has_strange(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this particle contain a strange quark?"""
    return _has_quark_q(pdgid, 3)


def has_charm(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this particle contain a charm quark?"""
    return _has_quark_q(pdgid, 4)


def has_bottom(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this particle contain a bottom quark?"""
    return _has_quark_q(pdgid, 5)


def has_top(pdgid):
    # type: (PDGID_TYPE) -> bool
    """Does this particle contain a top quark?"""
    return _has_quark_q(pdgid, 6)


def has_fundamental_anti(pdgid):
    # type: (PDGID_TYPE) -> bool
    """If this is a fundamental particle, does it have a valid antiparticle?"""
    # These are defined by the generator and therefore are always valid
    fid = _fundamental_id(pdgid)
    if fid in range(80, 101):
        return True
    # Check PDGIDs from 1 to 79
    _cp_conjugates = (21, 22, 23, 25, 32, 33, 35, 36, 39, 41)
    if fid in range(1, 80) and fid not in _cp_conjugates and is_valid(abs(int(pdgid))):
        return True
    return False


def charge(pdgid):
    # type: (PDGID_TYPE) -> Optional[float]
    """Returns the charge."""

    three_charge_pdgid = three_charge(pdgid)
    if three_charge_pdgid is None:
        return None
    elif not is_Qball(pdgid):
        return three_charge_pdgid / 3.0
    else:
        return three_charge_pdgid / 30.0


def three_charge(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """
    Returns 3 times the charge.

    None is returned if the PDGID is not valid.
    """
    if not is_valid(pdgid):
        return None

    aid = abspid(pdgid)
    charge = None
    ch100 = [
        -1,
        2,
        -1,
        2,
        -1,
        2,
        -1,
        2,
        0,
        0,
        -3,
        0,
        -3,
        0,
        -3,
        0,
        -3,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        0,
        0,
        3,
        0,
        0,
        0,
        0,
        -1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        6,
        3,
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]
    q1 = _digit(pdgid, Location.Nq1)
    q2 = _digit(pdgid, Location.Nq2)
    q3 = _digit(pdgid, Location.Nq3)
    sid = _fundamental_id(pdgid)

    if _extra_bits(pdgid) > 0:
        if is_nucleus(pdgid):  # ion
            Z_pdgid = Z(pdgid)
            if Z_pdgid is None:
                return None
            else:
                return 3 * Z_pdgid
        elif is_Qball(pdgid):  # Qball
            charge = 3 * ((aid // 10) % 10000)
        else:  # this should never be reached in the present numbering scheme
            return None  # since extra bits exist only for Q-balls and nuclei
    elif is_dyon(pdgid):  # Dyon
        charge = 3 * ((aid // 10) % 1000)
        # this is half right
        # the charge sign will be changed below if pid < 0
        if _digit(pdgid, Location.Nl) == 2:
            charge = -charge
    elif sid > 0 and sid <= 100:  # use table
        charge = ch100[sid - 1]
        if aid in (1000017, 1000018, 1000034, 1000052, 1000053, 1000054):
            charge = 0
        if aid == 5100061 or aid == 5100062:
            charge = 6
    elif _digit(pdgid, Location.Nj) == 0:  # KL, KS, or undefined
        return 0
    elif q1 == 0 or (is_Rhadron(pdgid) and q1 == 9):  # mesons
        if q2 == 3 or q2 == 5:
            charge = ch100[q3 - 1] - ch100[q2 - 1]
        else:
            charge = ch100[q2 - 1] - ch100[q3 - 1]
    elif q3 == 0:  # diquarks
        charge = ch100[q2 - 1] + ch100[q1 - 1]
    elif is_baryon(pdgid) or (
        is_Rhadron(pdgid) and _digit(pdgid, Location.Nl) == 9
    ):  # baryons
        charge = ch100[q3 - 1] + ch100[q2 - 1] + ch100[q1 - 1]

    if charge is not None and int(pdgid) < 0:
        charge = -charge
    return charge


def j_spin(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """Returns the total spin as 2J+1."""
    if not is_valid(pdgid):
        return None
    if _fundamental_id(pdgid) > 0:
        fund = _fundamental_id(pdgid)
        if fund > 0 and fund < 7:
            return 2  # 4th generation quarks not dealt with !
        if fund == 9:
            return 3
        if fund > 10 and fund < 17:
            return 2  # 4th generation leptons not dealt with !
        if fund > 20 and fund < 25:
            return 3
        return None
    elif abs(int(pdgid)) in (1000000010, 1000010010):  # neutron, proton
        return 2
    elif _extra_bits(pdgid) > 0:
        return None
    if pdgid in (130, 310):
        return 1  # Special cases of the KS and KL !
    return abspid(pdgid) % 10


def J(pdgid):
    # type: (PDGID_TYPE) -> Optional[float]
    """Returns the total spin J."""
    value = j_spin(pdgid)
    return (
        (value - 1) / 2 if value is not None else value
    )  # This works due to the Python 3 style division


def S(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """
    Returns the spin S.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    if not is_meson(pdgid):
        return None
    if not is_valid(pdgid):
        return None

    nl = (abspid(pdgid) // 10000) % 10
    js = abspid(pdgid) % 10
    if (abspid(pdgid) // 1000000) % 10 == 9:
        return None  # no knowledge so far
    if nl == 0 and js >= 3:
        return 1
    elif nl == 0 and js == 1:
        return 0
    elif nl == 1 and js >= 3:
        return 0
    elif nl == 2 and js >= 3:
        return 1
    elif nl == 1 and js == 1:
        return 1
    elif nl == 3 and js >= 3:
        return 1
    return 0


def s_spin(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """
    Returns the spin S as 2S+1.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    value = S(pdgid)
    return (2 * value + 1) if value is not None else value


def L(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """
    Returns the orbital angular momentum L.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    if not is_meson(pdgid):
        return None
    if not is_valid(pdgid):
        return None

    nl = (abspid(pdgid) // 10000) % 10
    js = abspid(pdgid) % 10
    if (abspid(pdgid) // 1000000) % 10 == 9:
        return None  # no knowledge so far

    if nl == 0 and js == 3:
        return 0
    elif nl == 0 and js == 5:
        return 1
    elif nl == 0 and js == 7:
        return 2
    elif nl == 0 and js == 9:
        return 3
    elif nl == 0 and js == 1:
        return 0
    elif nl == 1 and js == 3:
        return 1
    elif nl == 1 and js == 5:
        return 2
    elif nl == 1 and js == 7:
        return 3
    elif nl == 1 and js == 9:
        return 4
    elif nl == 2 and js == 3:
        return 1
    elif nl == 2 and js == 5:
        return 2
    elif nl == 2 and js == 7:
        return 3
    elif nl == 2 and js == 9:
        return 4
    elif nl == 1 and js == 1:
        return 1
    elif nl == 3 and js == 3:
        return 2
    elif nl == 3 and js == 5:
        return 3
    elif nl == 3 and js == 7:
        return 4
    elif nl == 3 and js == 9:
        return 5
    return 0


def l_spin(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """
    Returns the orbital angular momentum L as 2L+1.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    value = L(pdgid)
    return (2 * value + 1) if value is not None else value


def A(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """Returns the atomic number A if the PDG ID corresponds to a nucleus. Else it returns None."""
    # A proton can be a Hydrogen nucleus
    # A neutron can be considered as a nucleus when given the PDG ID 1000000010,
    # hence consistency demands that A(neutron) = 1
    if abspid(pdgid) in (2112, 2212):
        return 1
    if _digit(pdgid, Location.N10) != 1 or _digit(pdgid, Location.N9) != 0:
        return None
    return (abspid(pdgid) // 10) % 1000


def Z(pdgid):
    # type: (PDGID_TYPE) -> Optional[int]
    """Returns the charge Z if the PDG ID corresponds to a nucleus. Else it returns None."""
    # A proton can be a Hydrogen nucleus
    if abspid(pdgid) == 2212:
        return int(pdgid) // 2212
    # A neutron can be considered as a nucleus when given the PDG ID 1000000010,
    # hence consistency demands that Z(neutron) = 0
    if abspid(pdgid) == 2112:
        return 0
    if _digit(pdgid, Location.N10) != 1 or _digit(pdgid, Location.N9) != 0:
        return None
    return ((abspid(pdgid) // 10000) % 1000) * (int(pdgid) // abs(int(pdgid)))


def _digit(pdgid, loc):
    # type: (PDGID_TYPE, int) -> int
    """
    Provides a convenient index into the PDGID number, whose format is in base 10.

    Returns the digit at position 'loc' given that the right-most digit is at position 1.
    """
    sid = str(abspid(pdgid))
    return int(sid[-loc]) if loc <= len(sid) else 0


def _extra_bits(pdgid):
    # type: (PDGID_TYPE) -> int
    """
    Returns everything beyond the 7th digit, so anything outside the PDG numbering scheme.
    """
    return abspid(pdgid) // 10000000


def _fundamental_id(pdgid):
    # type: (PDGID_TYPE) -> int
    """
    Returns the first 2 digits if this is a "fundamental" particle.
    Returns 0 if the particle is not fundamental or not standard (PDG ID with more than 7 digits).

    PDGID=100 is a special case (internal generator ID's are 81-100).
    """
    if _extra_bits(pdgid) > 0:
        return 0
    if _digit(pdgid, Location.Nq2) == 0 and _digit(pdgid, Location.Nq1) == 0:
        return abspid(pdgid) % 10000
    elif abspid(pdgid) <= 100:
        return abspid(pdgid)
    else:
        return 0


def _has_quark_q(pdgid, q):
    # type: (PDGID_TYPE, int) -> bool
    """
    Helper function - does this particle contain a quark q?

    Note that q is always positive, so [1, 6] for Standard Model quarks.
    """
    # Nuclei can also contain strange quarks,
    # cf. the definition of a nucleus PDG ID in is_nucleus.
    # This check needs to be done first since _extra_bits(pdgid) > 0 for nuclei
    if is_nucleus(pdgid):
        if q in (1, 2):
            return True  # Nuclei by construction contain up and down quarks
        elif q == 3 and pdgid not in (2112, 2212):
            if _digit(pdgid, Location.N8) > 0:
                return True
            else:
                return False
    if _extra_bits(pdgid) > 0:
        return False
    if _fundamental_id(pdgid) > 0:
        return False
    if is_dyon(pdgid):
        return False
    if is_Rhadron(pdgid):
        iz = 7
        for loc in range(6, 1, -1):
            if _digit(pdgid, loc) == 0:
                iz = loc
            elif loc == iz - 1:
                # ignore squark or gluino
                pass
            else:
                if _digit(pdgid, loc) == q:
                    return True
        return False
    if (
        _digit(pdgid, Location.Nq3) == q
        or _digit(pdgid, Location.Nq2) == q
        or _digit(pdgid, Location.Nq1) == q
    ):
        return True
    if is_pentaquark(pdgid):
        if _digit(pdgid, Location.Nl) == q or _digit(pdgid, Location.Nr) == q:
            return True
    return False



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Class representing a Pythia ID.
"""

from __future__ import absolute_import

import csv

from .. import data
from ..pdgid import PDGID
from ..exceptions import MatchingIDNotFound


with data.open_text(data, "pdgid_to_pythiaid.csv") as _f:
    _bimap = {
        int(v["PYTHIAID"]): int(v["PDGID"])
        for v in csv.DictReader(l for l in _f if not l.startswith("#"))
    }


class PythiaID(int):
    """
    Holds a Pythia ID.

    Examples
    --------
    >>> pythiaid = PythiaID(211)

    >>> from particle import Particle
    >>> p = Particle.from_pdgid(pythiaid.to_pdgid())

    >>> p = Particle.find(pdgid=pythiaid.to_pdgid())
    >>> p.name
    'pi+'
    """

    __slots__ = ()  # Keep PythiaID a slots based class

    @classmethod
    def from_pdgid(cls, pdgid):
        """
        Constructor from a PDGID.
        """
        for k, v in _bimap.items():
            if v == pdgid:
                return cls(k)
        raise MatchingIDNotFound(
            "Non-existent PythiaID for input PDGID {0} !".format(pdgid)
        )

    def to_pdgid(self):
        return PDGID(_bimap[self])

    def __repr__(self):
        return "<PythiaID: {:d}>".format(int(self))

    def __str__(self):
        return repr(self)

    def __neg__(self):
        return self.__class__(-int(self))

    __invert__ = __neg__



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Class representing a PDG ID.

All methods of HepPID are implemented in a Pythonic version, see the functions module.
"""

from __future__ import absolute_import

from . import functions as _functions

from inspect import isfunction


# Collect all the user defined, non-hidden functions in the pdgid.functions module
_fnames = [
    fname
    for fname in dir(_functions)
    if not fname.startswith("_") and isfunction(getattr(_functions, fname))
]


class PDGID(int):
    """
    Holds a PDG ID.

    Example
    -------
    >>> PDGID(11).is_lepton
    True
    """

    __slots__ = ()  # Keep PDGID a slots based class

    def __repr__(self):
        # type: () -> str
        return "<PDGID: {:d}{:s}>".format(
            int(self), "" if _functions.is_valid(self) else " (is_valid==False)"
        )

    def __str__(self):
        # type: () -> str
        return repr(self)

    def __neg__(self):
        # type: () -> PDGID
        return self.__class__(-int(self))

    __invert__ = __neg__

    def info(self):
        # type: () -> str
        """
        Print all PDGID properties one per line, for easy inspection.
        """
        val = ""
        for item in _fnames:
            val += "{item:14} {value}\n".format(item=item, value=getattr(self, item))
        return val


# Decorate the PDGID class with all relevant functions defined in the pdgid.functions module
for _n in _fnames:
    _decorator = property(getattr(_functions, _n), doc=getattr(_functions, _n).__doc__)
    setattr(PDGID, _n, _decorator)



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import absolute_import

from .particle import Particle
from .particle import ParticleNotFound, InvalidParticle
from .enums import SpinType, Parity, Charge, Inv, Status
from .kinematics import width_to_lifetime, lifetime_to_width
from .utilities import latex_to_html_name
